"""
Copyright 2019 European Union

Licensed under the EUPL, Version 1.2 or as soon they will be approved by the European Commission  subsequent versions of the EUPL (the "Licence");

You may not use this work except in compliance with the Licence.
You may obtain a copy of the Licence at:

https://joinup.ec.europa.eu/sites/default/files/inline-files/EUPL%20v1_2%20EN(1).txt

Unless required by applicable law or agreed to in writing, software distributed under the Licence is distributed on an "AS IS" basis,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the Licence for the specific language governing permissions and limitations under the Licence.

cutmaps
A tool to cut netcdf files
"""

import argparse
import os
import shutil
import sys

from .cutlib import get_filelist, get_cuts, cutmap, logger


class ParserHelpOnError(argparse.ArgumentParser):
    def error(self, message):
        sys.stderr.write('Error: %s\n' % message)
        self.print_help()
        sys.exit(1)

    def add_args(self):
        group_mask = self.add_mutually_exclusive_group()
        group_filelist = self.add_mutually_exclusive_group()

        group_mask.add_argument("-m", "--mask", help='mask file cookie-cutter, .map if pcraster, .nc if netcdf')
        group_mask.add_argument("-c", "--cuts", help='Cut coordinates in the form lonmin_lonmax:latmin_latmax')

        group_filelist.add_argument("-l", "--list", help='list of files to be cut, in a text file, one file per line, '
                                                         'main variable i.e. pr/e0/tx/tavg must be '
                                                         'last variable in the input file')
        group_filelist.add_argument("-f", "--folder", help='Directory with netCDF files to be cut')
        group_filelist.add_argument("-g", "--global-setup", help='Directory with GloFAS static maps. '
                                                                 'Output files will have same directories structure')

        self.add_argument("-o", "--outpath", help='path where to save cut files', default='./cutmaps_out', required=True)


def main(cliargs):
    parser = ParserHelpOnError(description='Convert PCRaster maps to a NetCDF map stack')

    parser.add_args()

    args = parser.parse_args(cliargs)

    mask = args.mask
    cuts = args.cuts
    filelist = args.list

    input_folder = args.folder
    glofas_folder = args.global_setup
    pathout = args.outpath

    logger.info('\n\nCutting using: %s\n Files to cut from: %s\n Output: %s\n\n ',
                mask or cuts,
                filelist or input_folder or glofas_folder,
                pathout)

    list_to_cut = get_filelist(filelist, input_folder, glofas_folder)
    x_max, x_min, y_max, y_min = get_cuts(cuts, mask)

    # walk through list_to_cut
    for file_to_cut in list_to_cut:

        filename, ext = os.path.splitext(file_to_cut)

        # localdir used only with glofas_folder.
        # It will track folder structures in a GloFAS like setup and replicate it on output folder
        localdir = os.path.dirname(file_to_cut).replace(os.path.dirname(glofas_folder), '').lstrip('/') if glofas_folder else ''

        fileout = os.path.join(pathout, localdir, os.path.basename(file_to_cut))
        if os.path.isdir(file_to_cut) and glofas_folder:
            # just create folder
            os.makedirs(fileout, exist_ok=True)
            continue
        if ext != '.nc':
            if glofas_folder:
                logger.warning('%s is not in netcdf format, just copying to ouput folder', file_to_cut)
                shutil.copy(file_to_cut, fileout)
            else:
                logger.warning('%s is not in netcdf format, skipping...', file_to_cut)
            continue
        elif os.path.isfile(fileout) and os.path.exists(fileout):
            logger.warning('%s already existing. This file will not be overwritten', fileout)
            continue

        cutmap(file_to_cut, fileout, x_max, x_min, y_max, y_min)


def main_script():
    sys.exit(main(sys.argv[1:]))


if __name__ == '__main__':
    sys.exit(main(sys.argv[1:]))



import os
import logging
from pathlib import Path

import numpy as np
from netCDF4 import Dataset

logging.basicConfig(level=logging.INFO, format='[%(levelname)s] %(message)s')
logger = logging.getLogger()


class NetCDFComparator:
    def __init__(self, mask, atol=0.05, rtol=0.1, max_perc_diff=0.2, max_perc_large_diff=0.1):
        self.mask = Dataset(mask)
        maskvar = [k for k in self.mask.variables if len(self.mask.variables[k].dimensions) == 2][0]
        self.maskarea = np.logical_not(self.mask.variables[maskvar][:, :])
        self.atol = atol
        self.rtol = rtol
        self.max_perc_large_diff = max_perc_large_diff
        self.max_perc_diff = max_perc_diff
        self.large_diff_th = self.atol * 10
        self.errors = []

    def check_vars(self, maskarea, step, vara_step, varb_step, varname, filepath):
        step = step if step is not None else '(no time)'
        filepath = os.path.basename(filepath)
        vara_step = np.ma.masked_array(vara_step, maskarea)
        varb_step = np.ma.masked_array(varb_step, maskarea)
        vara_step = np.ma.compressed(vara_step).astype('float64')
        varb_step = np.ma.compressed(varb_step).astype('float64')
        diff_values = np.ma.abs(vara_step - varb_step)
        same_values = np.ma.allclose(diff_values, np.zeros(diff_values.shape), atol=self.atol, rtol=self.rtol)
        all_ok = vara_step.size == varb_step.size and same_values
        array_ok = np.isclose(diff_values, np.zeros(diff_values.shape), atol=self.atol, rtol=self.rtol, equal_nan=True)
        different_values_size = array_ok[~array_ok].size
        if (not all_ok) and (different_values_size > 0):
            max_diff = np.ma.amax(diff_values)
            perc_wrong = different_values_size * 100 / diff_values.size
            result = np.where(diff_values >= max_diff)
            rel_diff = max_diff * 100. / np.maximum(vara_step[result], varb_step[result])
            if np.max(rel_diff) > 0.01 and (perc_wrong >= self.max_perc_diff or (perc_wrong >= self.max_perc_large_diff and max_diff > self.large_diff_th)):
                rel_diff = np.array_str(rel_diff)
                logger.error(f'Var: {varname} - STEP {step}: {perc_wrong:3.2f}% of values are different. max diff: {max_diff:3.2f} (rel diff: {rel_diff}%)')
                logger.error('\nA: %s\nB: %s', np.array_str(vara_step[result]), np.array_str(varb_step[result]))
                return f'{filepath}/{varname}/{step} - {perc_wrong:3.2f}% of different values - max diff: {max_diff:3.2f} (rel diff: {rel_diff}%)'

    def compare_dirs(self, path_a, path_b):
        for fa in Path(path_a).glob('**/*.nc'):
            fb = os.path.join(path_b, os.path.basename(fa))
            if not os.path.exists(fb):
                logger.info('skipping %s as it is not in %s', fb, path_b)
                continue
            errors = self.compare_files(fa, fb)
            if errors:
                self.errors += errors
        return self.errors

    def compare_files(self, file_a, file_b):
        errors = []
        nca = Dataset(file_a)
        ncb = Dataset(file_b)
        num_dims = 3 if 'time' in nca.variables else 2
        var_name = [k for k in nca.variables if len(nca.variables[k].dimensions) == num_dims][0]
        vara = nca.variables[var_name]
        varb = ncb.variables[var_name]
        if 'time' in nca.variables:
            stepsa = nca.variables['time'][:]
            stepsb = ncb.variables['time'][:]
            if len(stepsa) != len(stepsb):
                logger.error('Different number of timesteps')
                len_stepsa = len(stepsa)
                len_stepsb = len(stepsb)
                return f'Files: {file_a} vs {file_b}: different size in time axis A:{len_stepsa} B:{len_stepsb}'
            for i, step in enumerate(stepsa):
                vara_step = vara[i][:, :]
                varb_step = varb[i][:, :]
                err = self.check_vars(self.maskarea, i, vara_step, varb_step, var_name, file_a)
                if err:
                    errors.append(err)
        else:
            vara_step = vara[:, :]
            varb_step = varb[:, :]
            err = self.check_vars(self.maskarea, None, vara_step, varb_step, var_name, file_a)
            if err:
                errors.append(err)
        nca.close()
        ncb.close()
        return errors
